//******************************************************************************
// ED Text Editor
// Copyright (c) 2011, Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <conio.h>
#include "smalloc.h"

#define EX_MODE

#define UNSPECIFIED   -2
#define PARSING_ERROR -3
#define TAB_CHAR      9
#define MAX_CHARS     200

#ifdef EX_MODE
#define PROMPT_CHAR  ':'
#define CURSOR_CHAR  '@'
#else
#define PROMPT_CHAR  '*'
#endif

#ifdef CSPIN_FLAG
#define MALLOC_SIZE 16000
#else
#define MALLOC_SIZE 80000
#endif

// Uninitialized global variables
int memorysize;
StringT *ListHead;
StringT *YankHead;
char srchstr[100];
char rplcstr1[100];
char rplcstr2[100];
char filename[100];
char inbuf[MAX_CHARS];

// Initialized global variables
int rflag = 0;
int gflag = 0;
int vmode = 1;
int prompt = 0;
int vlines = 21;
int scroll = 20;
int changed = 0;
int numlines = 0;
int currline = 0;
int currcol = 1;
int dispcol = 1;
#ifdef EX_MODE
int cmd0 = 0;
int count = 0;
int schange = 1;
int topline = 1;
#endif
char nullstr[1] = {0};

void CheckMemory(void);
int  ExecuteCommand(void);
void ExecuteViCommand(int cmd);
void PrintScreen(int linenum);
int  FindString(char *str1, char *str2);
int  ReadFile(char *fname, int linenum);

// Main program entry
#ifdef CSPIN_FLAG
int main(void)
{
    int argc = 1;
    char **argv;
#else
extern StringT *smfreelist;

int main(int argc, char **argv)
{
#endif
    int cmd;
    int retcode;
    char *strbuf;

    // Initialization
    srchstr[0] = 0;
    rplcstr1[0] = 0;
    rplcstr2[0] = 0;
    ListHead = 0;
    YankHead = 0;

#ifdef EX_MODE
    prompt = PROMPT_CHAR;
#endif

    // Allocate the string buffer
    for (memorysize = MALLOC_SIZE; memorysize > 0; memorysize -= 1000)
    {
        strbuf = malloc(memorysize);
	if (strbuf) break;
    }
    smallocinit(strbuf, memorysize);
    printf("memorysize = %d\n", memorysize);

    // Copy the input file name if specified
    if (argc > 1) strcpy(filename, argv[1]);
    else filename[0] = 0;

    currline = ReadFile(filename, 0);

#ifdef EX_MODE
    if (numlines > 0) currline = 1;
#endif

    // Get a command and execute it
    while (1)
    {
	gflag = 0;
	CheckMemory();
#ifdef EX_MODE
	if (vmode == 1)
	{
	    if (schange)
	    {
	        PrintScreen(currline);
		schange = 0;
	    }
	    cmd = getch();
	    ExecuteViCommand(cmd);
	}
	else
	{
	    if (prompt) printf("%c", prompt);
	    gets(inbuf);
	    retcode = ExecuteCommand();
	    if (retcode > 0)
	        break;
	    else if (retcode < 0)
                printf("E%d:\n", -retcode);
	    if (vmode == -1) vmode = 1;
	}
#else
	if (vmode) PrintScreen(currline);
	if (prompt) printf("%c", prompt);
	gets(inbuf);
	retcode = ExecuteCommand();
	if (retcode > 0)
	    break;
	else if (retcode < 0)
	    printf("?\n");
#endif
    }

    free(strbuf);
    return 0;
}

// Allocate a string link and copy contents of "str"
StringT *AllocateString(char *str)
{
    StringT *link = smalloc(strlen(str) + 1);
    if (link) strcpy(link->str, str);
    return link;
}

// Duplicate a string link
StringT *DuplicateString(StringT *link)
{
    return AllocateString(link->str);
}

// Check for a memory leak
void CheckMemory(void)
{
    int total;
    StringT *curr;
    int listsize = 0;
    int yanksize = 0;
    int freesize = 0;

    for (curr = ListHead; curr; curr = curr->next) listsize += curr->size;

    for (curr = YankHead; curr; curr = curr->next) yanksize += curr->size;

    for (curr = smfreelist; curr; curr = curr->next) freesize += curr->size;

    total = listsize + yanksize + freesize;

    if (total != memorysize)
    {
	printf("Memory leak!\n");
        printf("listsize = %d, yanksize = %d, freesize = %d, total = %d\n",
            listsize, yanksize, freesize, listsize+yanksize+freesize);
    }
}

// Get a pointer to a string link from its line number
StringT *GetLinePtr(int linenum, StringP *pprev)
{
    StringT *prev = 0;
    StringT *curr = ListHead;

    while (curr && --linenum > 0)
    {
	prev = curr;
	curr = curr->next;
    }

    *pprev = prev;
    return curr;
}

// Remove the CR and LF characters from the end of a line
void RemoveCRLF(char *str)
{
    int len = strlen(str);

    str += len - 1;

    while (len > 0)
    {
	if (*str != 10 && *str != 13) break;
	*str-- = 0;
	len--;
    }
}

// Insert a string link at the line given by "linenum"
void InsertLine(StringT *line, int linenum)
{
    StringT *prev;
    StringT *curr = GetLinePtr(linenum-1, &prev);

    if (linenum < 1 || linenum > numlines + 1)
        printf("InsertLine: linenum = %d, numlines = %d\n", linenum, numlines);

    if (curr == 0 && (linenum != 1 || ListHead != 0))
        printf("InsertLine: curr == 0, linenum = %d\n", linenum);

    if (linenum == 1)
    {
	line->next = ListHead;
	ListHead = line;
    }
    else
    {
	line->next = curr->next;
	curr->next = line;
    }

    numlines++;
}

// Delete the line at "linenum"
void DeleteLine(int linenum)
{
    StringT *prev;
    StringT *curr = GetLinePtr(linenum, &prev);
    if (!curr)
    {
	printf("DeleteLine: curr = 0, linenum = %d\n", linenum);
	return;
    }
    if (!prev) ListHead   = curr->next;
    else       prev->next = curr->next;
    sfree(curr);
    numlines--;
}

// Based on the value of gflag, check for if the line
// at "lineptr" does, or does not contains "str"
int CheckMatch(char *lineptr, char *str)
{
    if (gflag == 0)
        return 0;

    if (gflag == 1)
    {
	if (!FindString(lineptr, str)) return 1;
	return 0;
    }

    if (FindString(lineptr, str)) return 1;
    return 0;
}

// Print lines using the "n", "v", "l" or "p" format
int PrintLines(int linenum, int first, int last, int mode)
{
    int col;
    int len;
    int line;
    int fflag;
    int lastproc = linenum;
    char *LinePtr;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);

    for (line = first; line <= last; line++)
    {
        if (!curr)
        {
	    printf("PrintLines: curr = 0, line = %s\n", line);
	    return 0;
        }
	LinePtr = curr->str;
	curr = curr->next;
	if (CheckMatch(LinePtr, srchstr)) continue;
	if (mode == 'n')
	    printf("%6d  %s\n", line, LinePtr);
	else if (mode == 'v')
	{
#ifndef EX_MODE
	    if (line == linenum)
	        printf("%6d >%s\n", line, LinePtr);
	    else
	        printf("%6d  %s\n", line, LinePtr);
#else
	    if (line == linenum)
	    {
		len = strlen(LinePtr);
		for (col = 0; col < len; col++)
		{
		    if (col+1 != dispcol) printf("%c", LinePtr[col]);
		    else printf("%c", CURSOR_CHAR);
		}
		if (dispcol > len) printf("%c", CURSOR_CHAR);
		printf("\n");
	    }
	    else
	        printf("%s\n", LinePtr);
#endif
	}
	else if (mode == 'l')
	{
	    while (*LinePtr)
	    {
		if (*LinePtr >= 32 && *LinePtr < 127)
		    printf("%c", *LinePtr);
		else if (*LinePtr == TAB_CHAR)
		    printf("\\t");
		else
		    printf("\\x%2.2x", *LinePtr);
		LinePtr++;
	    }
	    printf("$\n");
	}
	else printf("%s\n", LinePtr);
	lastproc = line;
    }
    return lastproc;
}

// Decode a number from the string
int getnum(char **pstr)
{
    int val = 0;
    char *str = *pstr;

    while (isdigit(*str))
    {
	val = (val * 10) + *str++ - '0';
    }

    *pstr = str;
    return val;
}

// Find the next character that matches "val"
char *FindChar(char *str, int val)
{
    while (*str)
    {
	if (*str == val) break;
	str++;
    }
    return str;
}

// Skip the next character that does not match "val"
char *SkipChar(char *str, int val)
{
    while (*str)
    {
	if (*str != val) break;
	str++;
    }
    return str;
}

// Return true if "str2" is found in "str1"
int FindString(char *str1, char *str2)
{
    int len;
    int col = 1;

    len = strlen(str2);

    while (*str1)
    {
	if (strncmp(str1, str2, len) == 0) return col;
	str1++;
	col++;
    }
    return 0;
}

// Use the first character as a delimeter and copy the string to srchstr
int GetString(char **pstr)
{
    char *ptr;
    char *str = *pstr;
    int val = *str++;
    int len;

    if (!val) return 0;

    ptr = FindChar(str, val);
    len = ((int)ptr) - ((int)str);
    if (*ptr) ptr++;
    *pstr = ptr;
    if (len > 0)
    {
        memcpy(srchstr, str, len);
        srchstr[len] = 0;
    }
    return 1;
}

// Search forward through all the lines for the first occurance of srchstr
int SearchStringForward(char **pstr, int linenum)
{
    int i;
    StringT *prev;
    StringT *curr = GetLinePtr(linenum + 1, &prev);

    if (!GetString(pstr)) return PARSING_ERROR;

    for (i = linenum + 1; i <= numlines; i++)
    {
	if (!curr) { printf("SearchStringForward: curr = 0, line = %d\n", i); return -3;}
	if (FindString(curr->str, srchstr)) return i;
	curr = curr->next;
    }

    curr = ListHead;
    for (i = 1; i <= linenum; i++)
    {
	if (!curr) { printf("SearchStringForward: curr = 0, line = %d\n", i); return -3;}
	if (FindString(curr->str, srchstr)) return i;
	curr = curr->next;
    }

    return PARSING_ERROR;
}

// Search backward through all the lines for the first occurance of srchstr
// Note: This is not as efficient as the forward seach because we must use
// GetLinePtr to get the string link pointer
int SearchStringBackward(char **pstr, int linenum)
{
    int i;
    StringT *prev;
    StringT *curr;

    if (!GetString(pstr)) return PARSING_ERROR;

    for (i = linenum - 1; i >= 1; i--)
    {
        curr = GetLinePtr(i, &prev);
	if (!curr) { printf("SearchStringBackward: curr = 0, line = %d\n", i); return -3;}
        if (FindString(curr->str, srchstr)) return i;
    }

    for (i = numlines; i >= linenum; i--)
    {
        curr = GetLinePtr(i, &prev);
	if (!curr) { printf("SearchStringBackward: curr = 0, line = %d\n", i); return -3;}
        if (FindString(curr->str, srchstr)) return i;
    }

    return PARSING_ERROR;
}

// Parse the replace string command and extract rplcstr1 and rplcstr2
int ParseReplaceString(char *str)
{
    char *ptr;

    if (*str++ != 's') return 0;
    if (*str == 0) return 1;

    if (*str++ != '/') return 0;
    ptr = FindChar(str, '/');
    if (*ptr != '/') return 0;
    *ptr++ = 0;
    if (*str) strcpy(rplcstr1, str);
    else      strcpy(rplcstr1, srchstr);
    if (*ptr == 0) return 1;

    str = FindChar(ptr, '/');
    if (*str != '/') return 0;
    *str++ = 0;
    strcpy(rplcstr2, ptr);

    rflag = (*str == 'g');

    return 1;
}

// Store the results of the replace-string command on str1 into str2
int ReplaceString(char *str1, char *str2)
{
    int changeflag = 0;
    int len1 = strlen(rplcstr1);
    int len2 = strlen(rplcstr2);

    // Search through str1 for rplcstr1 and replace with rplcstr2
    while (*str1)
    {
	if (strncmp(str1, rplcstr1, len1) == 0)
	{
	    changeflag = 1;
	    memcpy(str2, rplcstr2, len2);
	    str2 += len2;
	    str1 += len1;
	    // Quit after the first time if 'g' was not specified
	    if (!rflag) break;
	}
	else
	    *str2++ = *str1++;
    }
    // Copy the remainder of str1 to str2
    while (*str1)
    {
        *str2++ = *str1++;
    }
    *str2 = 0;
    return changeflag;
}

// Execute the replace-string command on multiple lines in the text buffer
int ReplaceStringCommand(char *inptr, int first, int last, int linenum)
{
    int i;
    int lastproc = linenum;
    int changeflag = 0;
    StringT *link;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);

    if (!ParseReplaceString(inptr))
    {
	printf("?\n");
	return last;
    }

    for (i = first; i <= last; i++)
    {
        if (!curr) { printf("ReplaceStringCommand: line = %d\n", i); return 1; }

        if (CheckMatch(curr->str, srchstr))
	{
	    prev = curr;
	    curr = curr->next;
	    continue;
	}

        if(ReplaceString(curr->str, inbuf))
	{
	    link = curr;
            curr = smalloc(strlen(inbuf) + 1);
	    if (!curr) return lastproc;
	    strcpy(curr->str, inbuf);
            if (prev) prev->next = curr;
            else      ListHead   = curr;
            curr->next = link->next;
	    sfree(link);
	    changeflag = 1;
	    changed = 1;
	}
	prev = curr;
	curr = curr->next;
        lastproc = i;
    }

    if (!changeflag) printf("?\n");

    return lastproc;
}

// Extract a line number parameter from a command
char *GetLineNumber(char *str, int linenum, int *num)
{
    int val;

    if (str[0] == '.')
    {
	*num = linenum;
	str++;
    }
    else if (str[0] == '-')
    {
	str++;
	if (isdigit(*str))
	    val = getnum(&str);
	else
	    val = 1;
	*num = linenum - val;
    }
    else if (str[0] == '+')
    {
	str++;
	if (isdigit(*str))
	    val = getnum(&str);
	else
	    val = 1;
	*num = linenum + val;
    }
    else if (str[0] == '$')
    {
	*num = numlines;
	str++;
    }
    else if (str[0] == '/')
    {
	*num = SearchStringForward(&str, linenum);
    }
    else if (str[0] == '?')
    {
        *num = SearchStringBackward(&str, linenum);
    }
    else if (isdigit(*str))
    {
	*num = getnum(&str);
    }
    else *num = UNSPECIFIED;
    if (*num != UNSPECIFIED &&
        (*num < 0 || *num > numlines)) *num = PARSING_ERROR;
    return str;
}

// Extract the first and last numbers and the command from a command string
int ParseCommand(char **pstr, int linenum, int *first, int *last, int *cmd, int *bang)
{
    int done = 0;
    char *str = *pstr;

    *bang = 0;

#ifdef EX_MODE
    if (*str == '%')
#else
    if (*str == ',')
#endif
    {
	*first = 1;
	*last = numlines;
	done = 1;
    }
    else if (*str == ';')
    {
	*first = linenum;
	*last = numlines;
	done = 1;
    }
#ifdef EX_MODE
    else if (*str == ',')
    {
	*first = *last = numlines;
	done = 1;
    }
#endif

    if (done)
    {
	str++;
	*cmd = *str;
	if (*str && str[1] == '!')
	{
	    *bang = 1;
	    str++;
	}
	*pstr = str;
	return 1;
    }

    str = GetLineNumber(str, linenum, first);
    if (*first == PARSING_ERROR) return 0;
    if (*first == UNSPECIFIED)
    {
	if (numlines == 0 && *str == 'i')
	{
	    *first = *last = 1;
	}
	else if (*str == 'g' || *str == 'v' || *str == 'w' || *str == 'x')
	{
	    *first = 1;
	    *last = numlines;
	}
	else if (*str == 'j')
	{
	    *first = linenum;
	    *last = linenum + 1;
	    if (*last > numlines) return 0;
	}
	else if (*str == 'r')
	{
	    *first = numlines;
	    *last = numlines;
	}
	else if (*str == 0 || *str == 'z')
	{
	    *first = *last = linenum + 1;
	    if (*first > numlines) return 0;
	}
	else if (*str == '=')
	{
	    *first = *last = numlines;
	}
	else
	{
	    *first = linenum;
	    *last = linenum;
	}
    }
    else if (*str == ',' || *str == ';')
    {
	str = GetLineNumber(str+1, linenum, last);
        if (*last == PARSING_ERROR) return 0;
	if (*last == UNSPECIFIED) *last = *first;
	if (*last < (*first)) return 0;
    }
    else *last = *first;
    *cmd = *str;
    if (*str && str[1] == '!')
    {
        *bang = 1;
        str++;
    }
    *pstr = str;
    // If first is zero, check if commands allows that
    if (*first == 0)
    {
        str = FindChar("aq=EerxQhVfP", *cmd);
        if (*str == 0) return 0;
    }
    return 1;
}

#ifndef EX_MODE
// Print the help information
void help(void)
{
    printf("ED Commands\n");
    printf("\n");
    printf("Line Addresses\n");
    printf("n                - nth line\n");
    printf(".                - current line\n");
    printf("$                - last line\n");
    printf("-                - previous line\n");
    printf("-n               - nth previous line\n");
    printf("+                - next line\n");
    printf("+n               - nth next line\n");
    printf("n,m              - lines n through m\n");
    printf(",                - first through last lines\n");
    printf(";                - current through last lines\n");
    printf("/str/            - next line containing str\n");
    printf("?str?            - previous line containing str\n");
    printf("\n");
    printf("Editing Commands\n");
    printf("(.)a             - append text after the addressed line\n");
    printf("(.,.)c           - replace the addressed lines with text\n");
    printf("(.,.)d           - delete the addressed lines\n");
    printf("e file           - discard all lines and edit the specified file\n");
    printf("E file           - unconditionally discard all lines and edit specified file\n");
    printf("f file           - set the current file name and print the name\n");
    printf("(1,$)g/str/cmd   - execute a command on lines containing str\n");
    printf("h                - print help information\n");
    printf("(.)i             - insert text before the addressed line\n");
    printf("(.,.+1)j         - join the addressed lines\n");
    printf("(.,.)l           - print the addressed lines unambiguously\n");
    printf("(.,.)m(.)        - move the addressed lines to after the right-hand address\n");
    printf("(.,.)n           - print the addresses lines with line numbers\n");
    printf("(.,.)p           - print the addresses lines\n");
    printf("P                - toggle the prompt mode\n");
    printf("q                - quit\n");
    printf("($)r file        - read the specified file after the addressed line\n");
    printf("(.,.)s/st1/st2/g - substitute occurances of st1 with st2\n");
    printf("(.,.)t(.)        - copy the addressed lines to after the right-hand address\n");
    printf("(1,$)v/str/cmd   - execute a command on lines not containing str\n");
    printf("Vn               - set visualization mode to n lines or toggle on/off if no n\n");
    printf("(1,$)w file      - write the addressed lines to the specified file\n");
    printf("(.)x             - insert lines from the cut buffer after the addressed line\n");
    printf("(.,.)y           - yank the addressed lines to the cut buffer\n");
    printf("(.+1)zn          - scroll and print n lines from the addressed line\n");
    printf("($)=             - print the line number of the addressed line\n");
    printf("(.+1)            - print the addressed line\n");
}
#endif

// Print a visualization mode screen
#ifndef EX_MODE
void PrintScreen(int linenum)
{
    int i;
    int vlines1 = vlines - 1;
    int vlines2 = vlines/2;
    int first = linenum - vlines2;
    int last = first + vlines1;

    // Determine the first and last lines to display
    if (first < 1)
    {
	first = 1;
	last = first + vlines1;
        if (last > numlines) last = numlines;
    }
    else if (last > numlines)
    {
	last = numlines;
	first = last - vlines1;
	if (first < 1) first = 1;
    }

    printf("===============================================================\n");
    PrintLines(linenum, first, last, 'v');
    for (i = last - first; i < vlines1; i++) printf("~\n");
    printf("===============================================================\n");
}
#else
void PrintScreen(int linenum)
{
    int i;
    int vlines1 = vlines - 1;
    int vlines2 = vlines/2;
    int first = topline;
    int last = first + vlines1;

    // Clear the screen
    for (i = 0; i < 50; i++) printf("\n");

    // Determine the first and last lines to display
    if (first > currline)
    {
	first = currline;
	last = first + vlines1;
    }
    else if (currline > last)
    {
	last = currline;
	first = last - vlines1;
	if (first < 1) first = 1;
    }

    if (last > numlines) last = numlines;

    if (first > 0 && numlines > 0)
    {
        topline = first;
        PrintLines(linenum, first, last, 'v');
    }
    else last = first - 1;
    for (i = last - first; i < vlines1; i++) printf("~\n");
}
#endif

// Empty a list and free the string links
void EmptyList(StringT *list)
{
    StringT *next;

    while (list)
    {
	next = list->next;
	sfree(list);
	list = next;
    }
}

// Delete lines from the text file list and move them to the yank list
int DeleteLines(int first, int last)
{
    int i;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);
    StringT *yank = 0;

    EmptyList(YankHead);
    YankHead = 0;

    for (i = first; i <= last; i++)
    {
        if (!curr) { printf("DeleteLines: line = %d\n", first); return 1; }
	if (!yank) YankHead = curr;
	else       yank->next = curr;
	yank = curr;
	curr = curr->next;
	numlines--;
    }
    if (prev) prev->next = curr;
    else      ListHead   = curr;
    yank->next = 0;
    changed = 1;
    if (first > numlines) first = numlines;
    return first;
}

// Copy lines from the text file list to the yank list
int YankLines(int first, int last)
{
    int line;
    StringT *link;
    StringT *list = 0;
    StringT *curr = GetLinePtr(first, &link);

    EmptyList(YankHead);
    YankHead = 0;

    for (line = first; line <= last; line++)
    {
        if (!curr) { printf("YankLines: line = %d\n", line); return 1; }
	link = DuplicateString(curr);
	if (!link) break;
	if (!list) YankHead = link;
	else       list->next = link;
	curr = curr->next;
	list = link;
    }

    if (list) list->next = 0;
    return last;
}

// Copy the lines on the yank list to the text file list
int PushLines(int line)
{
    StringT *link;
    StringT *curr = YankHead;

    while (curr)
    {
	link = DuplicateString(curr);
	if (!link) break;
	InsertLine(link, ++line);
	curr = curr->next;
	changed = 1;
    }

    return line;
}

// Go into the text input mode and append after "linenum"
int AppendLines(int linenum)
{
    StringT *prev;
    StringT *curr = GetLinePtr(linenum+1, &prev);
    StringT *link;

    while (1)
    {
#ifdef CSPIN_FLAG
        gets(inbuf);
#else
        fgets(inbuf, MAX_CHARS, stdin);
#endif
        RemoveCRLF(inbuf);
        if (inbuf[0] == '.' && inbuf[1] == 0) break;
	link = AllocateString(inbuf);
	if (!link) break;
	link->next = curr;
	if (prev) prev->next = link;
	else      ListHead = link;
	prev = link;
	linenum++;
	numlines++;
	changed = 1;
    }
    return linenum;
}

// Go into the text input mode and insert lines before "linenum"
int InsertLines(int linenum)
{
    int linenum1 = AppendLines(linenum - 1);
    if (linenum1 > linenum) linenum = linenum1;
    return linenum;
}

// Join lines together and move the original lines to the yank buffer
int JoinLines(int first, int last)
{
    int i;
    int len = 0;
    char *str;
    StringT *prev;
    StringT *curr = GetLinePtr(first, &prev);
    StringT *link = curr;
    StringT *yank = 0;

    EmptyList(YankHead);
    YankHead = 0;

    // Determine the total length of the selected lines
    for (i = first; i <= last; i++)
    {
        if (!link) { printf("JoinLines: line = %d\n", i); return 1; }
	len += strlen(link->str);
	link = link->next;
    }

    // Allocate a new string buffer and concatenate the strings
    link = smalloc(len + 1);
    if (!link) return first;
    str = link->str;
    *str = 0;
    for (i = first; i <= last; i++)
    {
        if (!curr) { printf("JoinLines: line = %d\n", i); return 1; }
	strcat(str, curr->str);
	if (!yank) YankHead = curr;
	else       yank->next = curr;
	yank = curr;
	curr = curr->next;
	numlines--;
	changed = 1;
    }

    // Insert the new string
    if (prev) prev->next = link;
    else      ListHead   = link;
    link->next = curr;
    yank->next = 0;
    numlines++;

    return first;
}

// Move or copy lines to after the destination address
int MoveLines(int first, int last, int dstaddr, int cmd)
{
    int i;
    StringT *prev;
    StringT *curr;
    StringT *link;

    int srcaddr, insertbefore;

    if (dstaddr < 0 || dstaddr > numlines ||
        (first <= dstaddr && dstaddr < last))
    {
        printf("?\n");
    }
    else
    {
        srcaddr = first;
        insertbefore = dstaddr < first;
        for (i = first; i <= last; i++)
        {
            curr = GetLinePtr(srcaddr++, &prev);
            link = DuplicateString(curr);
            if (!link) break;
            InsertLine(link, ++dstaddr);
            if (insertbefore) srcaddr++;
            if (cmd == 'm')
            {
	        DeleteLine(--srcaddr);
	        if (!insertbefore) dstaddr--;
            }
        }
        changed = 1;
    }
    return dstaddr;
}

// Read a file after "linenum"
int ReadFile(char *fname, int linenum)
{
    int numchar = 0;
    FILE *infile = fopen(fname, "r");
    StringT *link;

    if (infile)
    {
        while (fgets(inbuf, MAX_CHARS, infile))
        {
	    RemoveCRLF(inbuf);
	    link = AllocateString(inbuf);
	    if (!link) break;
	    InsertLine(link, ++linenum);
            numchar += strlen(inbuf);
        }
        fclose(infile);
    }

    printf("%d\n", numchar);

    return linenum;
}

// Write lines first to last to the output file
int WriteFile(char *fname, int first, int last)
{
    int i;
    char *LinePtr;
    StringT *curr;
    StringT *prev;
    int numchar = 0;
    FILE *outfile = fopen(fname, "w");

    if (!outfile) return -1;

    for (i = first; i <= last; i++)
    {
        curr = GetLinePtr(i, &prev);
        LinePtr = curr->str;
        numchar += strlen(LinePtr);
        fprintf(outfile, "%s\n", LinePtr);
    }
    fclose(outfile);
    printf("%d\n", numchar);
    if (first == 1 && last == numlines) changed = 0;
    return 0;
}

// Execute a command
int ExecuteCommand(void)
{
    StringT *curr;
    StringT *prev;
    char *LinePtr;
    FILE *outfile;
    char *inptr = inbuf;
    int first, last, cmd, i, temp, numchar, bang;

    // Parse the command and extract the parameters
    if (!ParseCommand(&inptr, currline, &first, &last, &cmd, &bang)) return -2;

    // Parse a "g" or "v" command if present
    if (cmd == 'g' || cmd == 'v')
    {
        inptr++;
        if (GetString(&inptr))
        {
	    if (cmd == 'g')
	        gflag = 1;
	    else
	        gflag = 2;
	    cmd = *inptr;
	    if (cmd == 0) cmd = 'p';
	    printf("cmd = %c\n", cmd);
        }
        else return -3;
    }

    // Select the command and execute it
    switch (cmd)
    {
        case 0:
            // Print the current line if no command
            if (numlines == 0) return -4;
            if (last >= 0) currline = last;
            else if (first >= 0) currline = first;
            curr = GetLinePtr(currline, &prev);
            LinePtr = curr->str;
            printf("%s\n", LinePtr);
            break;
#ifdef EX_MODE
        case 'q':
	    if (changed == 0 || bang) return 1;
	    return -5;
	case 'e':
	    if (changed && bang == 0) return -6;
#else
        case 'Q':
	    // Quit unconditionally
            return 1;
        case 'q':
            // Check for changes and quit if OK
            if (changed == 0) return 1;
            changed = 0;
	    return -5;
#ifdef CSPIN_FLAG
        case 'e', 'E':
#else
	case 'E':
	case 'e':
#endif
	    // Reset the text buffer to the contents of the default file
	    if (changed && cmd == 'e')
	    {
	        changed = 0;
	        return -6;
	    }
#endif
	    inptr = SkipChar(inptr + 1, ' ');
	    if (*inptr == 0 && *filename == 0) return -7;
	    if (*inptr) strcpy(filename, inptr);
            EmptyList(YankHead);
            YankHead = 0;
            EmptyList(ListHead);
            ListHead = 0;
            numlines = 0;
            currline = ReadFile(filename, 0);
	    changed = 0;
	    break;
	case 'r':
	    // Read the contents of a file after the specified line
	    inptr = SkipChar(inptr + 1, ' ');
	    if (*inptr == 0 && *filename == 0) return -8;
            if (*inptr)
            {
                if (!filename[0]) strcpy(filename, inptr);
            }
            else inptr = filename;
	    temp = numlines;
            currline = ReadFile(inptr, first);
	    if (temp && numlines != temp) changed = 1;
	    break;
#ifdef EX_MODE
#ifdef CSPIN_FLAG
        case 'x', 'w':
#else
	case 'x':
	case 'w':
	    if (cmd == 'x' && changed == 0) return 1;
#endif
#else
	case 'w':
#endif
	    // Write the text buffer to a file
            inptr = SkipChar(inptr + 1, ' ');
	    if (*inptr == 0)
	    {
		if (*filename == 0) return -9;
		inptr = filename;
	    }
	    if (WriteFile(inptr, first, last) < 0) return -10;
	    if (cmd == 'x') return 1;
	    if (*filename == 0) strcpy(filename, inptr);
	    break;
	case 'd':
	    // Delete the specified lines
	    currline = DeleteLines(first, last);
	    break;
#ifdef CSPIN_FLAG
        case 'm', 't':
#else
	case 'm':
	case 't':
#endif
	    // Move or copy the specified lines
            if (inptr[1])
                sscanf(inptr + 1, "%d", &temp);
            else
                temp = currline;
            currline = MoveLines(first, last, temp, cmd);
	    break;
	case 'c':
	    // Change the specified lines
            temp = (last == numlines);
            currline = DeleteLines(first, last);
	    if (temp)
                currline = AppendLines(currline);
	    else
                currline = InsertLines(currline);
	    break;
	case 'a':
	    // Append text after the specified line
            currline = AppendLines(first);
	    break;
	case 'i':
	    // Insert text before the specified line
            currline = InsertLines(first);
	    break;
	case 'j':
	    // Join together lines
            currline = JoinLines(first, last);
	    break;
#ifdef EX_MODE
#ifdef CSPIN_FLAG
        case 'p', 'l', '#':
#else
	case 'p':
	case 'l':
	case '#':
	    if (cmd == '#') cmd = 'n';
#endif
#else
#ifdef CSPIN_FLAG
        case 'p', 'l', 'n':
#else
	case 'p':
	case 'l':
	case 'n':
#endif
#endif
	    // Print lines in either the "p", "l" or "n" format
            currline = PrintLines(currline, first, last, cmd);
	    break;
#ifndef EX_MODE
	    case 'h':
	    // Print help information
            help();
	    break;
#endif
	case 'V':
	    // Set the number of lines in the visualization mode or
	    // toggle the visualization mode if no lines are specified
            if (inptr[1])
            {
                 sscanf(inptr+1, "%d", &temp);
                if (temp <= 0) return -11;
                vlines = temp;
                vmode = 1;
            }
            else
#ifdef EX_MODE
                vmode = 1;
#else
                vmode = vmode ^ 1;
#endif
	    break;
	case 's':
	    // Replace text in the specified lines
	        currline = ReplaceStringCommand(inptr, first, last, currline);
		break;
#ifdef EX_MODE
        case 'P':
#else
	case 'x':
#endif
	    // Push text from the yank buffer
	        currline = PushLines(first);
		break;
	case 'y':
	    // Yank lines to the yank buffer
	        currline = YankLines(first, last);
		break;
#ifndef EX_MODE
	case 'z':
	    // Set the number of lines to scroll if specified,
	    // and scroll forward
            first = last;
            if (inptr[1])
            {
                sscanf(inptr + 1, "%d", &temp);
	        if (temp <= 0) return -12;
	        scroll = temp;
            }
            last = first + scroll;
            if (last > numlines) last = numlines;
            PrintLines(currline, first, last, 'p');
            currline = last;
	    break;
	case 'f':
	    // Set the default file name if specifed and print it
            inptr = SkipChar(inptr + 1, ' ');
            if (*inptr == 0 && *filename == 0) return -13;
            if (*inptr) strcpy(filename, inptr);
            printf("%s\n", filename);
	    break;
#endif
	case '=':
	    // Print the number of lines, or the line number if specified
            printf("%d\n", last);
	    break;
#ifndef EX_MODE
	case 'P':
	    // Toggle the prompt mode
	    if (prompt) prompt = 0;
	    else        prompt = PROMPT_CHAR;
	    break;
#endif
	default:
	    // Return error code if not a valid command
            return -14;
    }
    return 0;
}

#ifdef EX_MODE
void SetDispCol(void)
{
    int len;
    StringT *curr = GetLinePtr(currline, &curr);

    if (curr)
    {
        len = strlen(curr->str);
        if (currcol > len) dispcol = len;
        else dispcol = currcol;
        if (dispcol < 1) dispcol = 1;
    }
}

void ExecuteViCommand(int cmd)
{
    int temp2, index, val;
    char *inptr = inbuf;
    int temp = currline;
    StringT *link;
    int i, len;
    StringT *curr;

    switch (cmd)
    {
	case 'Q':
	    vmode = 0;
	    count = 0;
	    schange = 1;
	    prompt = ':';
	    break;
	case 'Y':
            if (count <= 1)
                temp2 = currline;
            else
                temp2 = currline + count - 1;
            if (temp2 > numlines) temp2 = numlines;
            YankLines(currline, temp2);
            count = 0;
	    break;
	case 'p':
            currline = PushLines(currline);
	    schange = (currline != temp);
            count = 0;
	    break;
	case ':':
	    vmode = -1;
	    count = 0;
	    schange = 1;
	    prompt = cmd;
	    break;
	case 'n':
	    strcpy(inptr, "//");
	    temp2 = SearchStringForward(&inptr, currline);
	    if (temp2 > 0) currline = temp2;
	    schange = (currline != temp);
	    break;
	case 'N':
	    strcpy(inptr, "//");
	    temp2 = SearchStringBackward(&inptr, currline);
	    if (temp2 > 0) currline = temp2;
	    schange = (currline != temp);
	    break;
#ifdef CSPIN_FLAG
	case 13, 'j':
#else
        case 13:
	case 'j':
#endif
	    if (currline < numlines)
	    {
	        currline++;
	        schange = 1;
                curr = GetLinePtr(currline, &curr);
		if (curr)
		{
		    len = strlen(curr->str);
		    if (currcol > len) dispcol = len;
		    else dispcol = currcol;
		    if (dispcol < 1) dispcol = 1;
		}
	    }
	    count = 0;
	    break;
	case 'k':
	    if (currline > 1)
	    {
	        currline--;
	        schange = 1;
                curr = GetLinePtr(currline, &curr);
		if (curr)
		{
		    len = strlen(curr->str);
		    if (currcol > len) dispcol = len;
		    else dispcol = currcol;
		    if (dispcol < 1) dispcol = 1;
		}
	    }
	    count = 0;
	    break;
	case 'h':
            if (dispcol > 1)
	    {
	        dispcol--;
		currcol = dispcol;
	        schange = 1;
	    }
	    count = 0;
	    break;
	case 'l':
            curr = GetLinePtr(currline, &curr);
	    if (curr) len = strlen(curr->str);
	    else len = 0;
            if (dispcol < len)
	    {
	        dispcol++;
		currcol = dispcol;
	        schange = 1;
	    }
	    count = 0;
	    break;
	case 6:
            // ^F
	    currline += vlines;
	    if (currline > numlines) currline = numlines;
	    schange = (currline != temp);
	    SetDispCol();
	    break;
	case 2:
            // ^B
	    currline -= vlines;
	    if (currline < 1) currline = 1;
	    schange = (currline != temp);
	    SetDispCol();
	    break;
	case 4:
            // ^D
	    currline += vlines/2;
	    if (currline > numlines) currline = numlines;
	    schange = (currline != temp);
	    SetDispCol();
	    break;
	case 21:
            // ^U
	    currline -= vlines/2;
	    if (currline < 1) currline = 1;
	    schange = (currline != temp);
	    SetDispCol();
	    break;
	case 'G':
	    if (count >= 1 && count <= numlines) currline = count;
	    else if (count == 0) currline = numlines;
	    schange = (currline != temp);
	    count = 0;
	    SetDispCol();
	    break;
	case 'J':
            if (count <= 1)
                temp2 = currline + 1;
            else
                temp2 = currline + count;
            if (temp2 > numlines) temp2 = numlines;
            currline = JoinLines(currline, temp2);
            count = 0;
            cmd0 = 0;
            schange = 1;
	    break;
	case 'd':
	    if (cmd0 == 'd')
	    {
	        if (count <= 1)
	            temp2 = currline;
	        else
	            temp2 = currline + count - 1;
	        if (temp2 > numlines) temp2 = numlines;
	        currline = DeleteLines(currline, temp2);
	        count = 0;
	        cmd0 = 0;
	        schange = 1;
	        SetDispCol();
	        return;
	    }
	    break;
	case 'x':
	    if (curr = GetLinePtr(currline, &curr))
	    {
	        inptr = curr->str;
	        len = strlen(inptr);
	        if (len)
	        {
		    index = dispcol - 1;
		    for (index = dispcol - 1; index < len; index++)
		        inptr[index] = inptr[index+1];
		    schange = 1;
		    len--;
		    if (dispcol > len && len > 0) dispcol = len;
		    currcol = dispcol;
	        }
	    }
	    break;
	case 'r':
	    if (curr = GetLinePtr(currline, &curr))
	    {
	        inptr = curr->str;
	        len = strlen(inptr);
                val = getch();
	        if (val != 27)
	        {
		    if (len)
		    {
		        index = dispcol - 1;
		        inptr[index] = val;
		    }
	            else
	            {
	                link = AllocateString(" ");
		        link->str[0] = val;
	                InsertLine(link, currline);
		        DeleteLine(currline+1);
	            }
	            currcol = dispcol;
	        }
	    }
	    break;
#ifdef CSPIN_FLAG
        case 'O', 'A', 'I', 'i', 'a':
#else
	case 'O':
	case 'A':
	case 'I':
	case 'i':
	case 'a':
#endif
            changed = 1;
	    if (numlines == 0)
	    {
	        index = 0;
	        inbuf[0] = 0;
	        len = 0;
	        currline = 1;
	    }
	    else if (cmd == 'O')
	    {
	        index = 0;
	        inbuf[0] = 0;
	        len = 0;
	    }
	    else
	    {
	        link = GetLinePtr(currline, &link);
	        strcpy(inbuf, link->str);
	        len = strlen(inbuf);
	        if (cmd == 'A') index = len;
	        else
	        {
		    if (len == 0 || cmd == 'I')
		    {
		        index = 0;
		    }
		    else
		    {
		        index = dispcol;
		        if (cmd == 'i') index--;
		    }
	        }
	        DeleteLine(currline);
	    }
            currcol = dispcol = index + 1;
	    while (1)
	    {
	        link = AllocateString(inbuf);
	        InsertLine(link, currline);
                PrintScreen(currline);
	        val = getch();
	        if (val == 27)
	        {
		    if (len > 0 && index >= len)
		    {
		        index--;
		        schange = 1;
	                dispcol = currcol = index + 1;
		    }
	            break;
	        }
	        if (val == 8)
	        {
		    if (index > 0)
		    {
		        index--;
		        for (i = index; i < len; i++)
		            inbuf[i] = inbuf[i+1];
		        len--;
		    }
	            DeleteLine(currline);
	        }
	        else if (val == 13)
	        {
		    if (index == 0)
		    {
	                link = AllocateString(nullstr);
	                InsertLine(link, currline);
		        currline++;
	                DeleteLine(currline);
		    }
		    else
		    {
	                for (i = len+1; i > index; i--)
		            inbuf[i] = inbuf[i-1];
	                inbuf[index++] = 0;
		        len++;
	                link = AllocateString(inbuf);
	                InsertLine(link, currline);
		        currline++;
	                DeleteLine(currline);
		        len -= index;
		        for (i = 0; i <= len; i++)
		            inbuf[i] = inbuf[i+index];
		        index = 0;
		    }
	        }
	        else
	        {
	            for (i = len+1; i > index; i--)
		        inbuf[i] = inbuf[i-1];
	            inbuf[index++] = val;
		    len++;
	            DeleteLine(currline);
	        }
	        dispcol = currcol = index + 1;
	    }
	    break;
	case '$':
            curr = GetLinePtr(currline, &curr);
	    if (curr)
	    {
	        len = strlen(curr->str);
	        if (!len) len++;
	        if (len != dispcol)
	        {
		    dispcol = len;
	            currcol = dispcol;
		    schange = 1;
	        }
	    }
	    count = 0;
	    break;
	default:
            if (isdigit(cmd))
            {
	        if (count == 0 && cmd == '0')
	        {
	            if (dispcol != 1)
	            {
	                dispcol = 1;
	                currcol = dispcol;
	                schange = 1;
	            }
	        }
	        count = (count * 10) + cmd - '0';
            }
	    break;
    }
    cmd0 = cmd;
}
#endif

#ifdef CSPIN_FLAG
#include "smalloc.c"
#endif
/*
+----------------------------------------------------------------------------+
|                          TERMS OF USE: MIT License                         |
+----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy|
|of this software and associated documentation files (the "Software"), to    |
|deal in the Software without restriction, including without limitation the  |
|rights to use, copy, modify, merge, publish, distribute, sublicense, and/or |
|sell copies of the Software, and to permit persons to whom the Software is  |
|furnished to do so, subject to the following conditions:                    |
|                                                                            |
|The above copyright notice and this permission notice shall be included in  |
|all copies or substantial portions of the Software.                         |
|                                                                            |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     |
|FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS|
|IN THE SOFTWARE.                                                            |
+----------------------------------------------------------------------------+
*/
